---- Mod Mangement
-- Only for internal version
--
-- Always check for presence before use.
--
-- @author Oskar Eisemuth
-- @copyright 2018-2020
-- @module commonapi._modmm

local modmm = {
}

require "serialize"

local _utils = commonapi.utils
local _log = _utils.logPrep("commonapi._modmm")
local _native = commonapi._native

function modmm.getDirFor(name)
	local userpath = _native.paths.getUserPath()
	if (userpath == nil) then
		_log("No UserPath found?")
		return
	end
	local path = userpath .. '/commonapi/'..name
	_native.fs.mkdir(path)
	return path
end


function modmm.getDirLocalMods()
	local userpath = _native.paths.getUserPath()
	if (userpath == nil) then
		_log("No UserPath found?")
		return
	end
	local path = userpath .. '/mods/'
	_native.fs.mkdir(path)
	return path
end

function modmm.getDirGameMods()
	local userpath = _native.paths.getGamePath()
	if (userpath == nil) then
		_log("No getGamePath found?")
		return
	end
	local path = userpath .. '/mods/'
	_native.fs.mkdir(path)
	return path
end


function modmm.download(filename, url)
	local pcmd
	local q_file = _utils.quoteParam(filename)
	local q_url = _utils.quoteParam(url)


	if (commonapi._info.sys == 1) then
		pcmd = 'curl -f  -w "HTTP Status: %{http_code}" -sS -m 10 --create-dirs -z '..q_file..' -o '..q_file..' '..q_url
	else
		--pcmd = '/usr/bin/curl --create-dirs -z '..q_repolocalfile..' -o '..q_repolocalfile..' '..q_repositoryurl
		pcmd = 'LD_LIBRARY_PATH=; LD_PRELOAD=; /usr/bin/curl -f  -w "HTTP Status: %{http_code}" -sS -m 10 --create-dirs -z '..q_file..' -o '..q_file..' '..q_url ..' 2>&1'
	end

	_log("Trying to run "..pcmd)
	local p = io.popen(pcmd)
	if (p ~= nil) then
		local msg = p:read("*all")
		print(msg)
		p:close()
		local statuscode = msg.match(msg, "HTTP Status: (%d+)")
		if (statuscode ~= nil and (statuscode == "200" or statuscode == "304")) then
			--
		else
			return false, msg
		end
	end
	return true
end

local test = { abc = "test" }

function modmm.downloadRepository(filename, repositoryurl)
	--local repositoryurl = "https://www.transportfever.net/filebase/repos/tpf2.json"
	local path = modmm.getDirFor("repos")
	if (path == nil) then
		return false
	end
	local repolocalfile = path.."/"..filename

	local ok, errormsg = modmm.download(repolocalfile, repositoryurl)

	if (ok == false) then
		return false, errormsg
	end

	ok, errormsg = modmm.convertRepositoryFilename(filename)
	if (ok == false) then
		return false, errormsg
	end

	return true
end




local function compareversion(a, b)
	local l = math.max(#a.version, #b.version)
	for i = 1, l do
		local ai = a.version[i] or 0
		local bi = b.version[i] or 0
		if (ai < bi) then
			return false
		end
		if (ai > bi) then
			return true
		end
	end
	return false
end

function modmm._repoConverterToModIdMap(files)
	local mapidtoversion = {}
	for m, fileentry in ipairs(files) do
		if (fileentry.modid ~= nil) then
			local feversion = _utils.convertVersion(fileentry.version)
			fileentry.version = feversion

			if (mapidtoversion[fileentry.modid] == nil) then
				mapidtoversion[fileentry.modid] = {}
			end
			table.insert(mapidtoversion[fileentry.modid], fileentry)
		end
	end

	for m, entry in pairs(mapidtoversion) do
		table.sort(entry, compareversion)
	end
	return mapidtoversion
end


function modmm.convertRepositoryFilename(filename)
	local dkjson = require("commonapi2.extlib.dkjson")
	local jsonstr

	local path = modmm.getDirFor("repos")
	if (path == nil) then
		return
	end
	local fullfilename = path..'/'..filename
	local jsonfile = io.open(fullfilename, "r")
	if (jsonfile == nil) then
		return false,  "Can't open"..fullfilename
	end

	jsonstr = jsonfile:read("*a")
	jsonfile:close()


	local repo, pos, err = dkjson.decode(jsonstr, 1, nil, nil)
	if err then
		print ("Error:", err)
		return false, err
	end

	if (type(repo.files) == "table") then
		local mapidtoversion = modmm._repoConverterToModIdMap(repo.files)

		local data = {
			repo = repo.repo,
			mods = mapidtoversion,
		}

		local marshalfile = io.open(fullfilename..'.m', "wb")
		if (marshalfile == nil) then
			return false, "Can't open"..fullfilename..'.m'
		end
		marshalfile:write(_native.marshal.encode(data))
		marshalfile:close()
	end

	return true
end


function modmm.applyRepositoryFilename(filename, mods)
	--local starttime = os.clock()

	local path = modmm.getDirFor("repos")
	if (path == nil) then
		return
	end
	local fullfilename = path..'/'..filename

	local marshalfile = io.open(fullfilename..'.m', "rb")
	if (marshalfile == nil) then
		return
	end
	local marshaldata = marshalfile:read('*a')
	marshalfile:close()
	local repo = _native.marshal.decode(marshaldata)

	local repomods = repo.mods or {}
	local repodata = repo.repo or {}

	for m, entry in ipairs(mods) do
		if (repomods[entry.id] ~= nil) then
			local repomodentries = repomods[entry.id]
			for x, repomodentry in ipairs(repomodentries) do
				repomodentry.repodata = repodata
			end
			local upgradeversion = repomodentries[1].version
			entry.repomods = repomodentries
			entry.repodata = repodata
			if (_utils.compareVersion(entry.version, upgradeversion) < 0) then
				entry.upgradeversion = upgradeversion
			end
		end
	end

	--print(string.format("elapsed time: %.2f\n", os.clock() - starttime))
	return repodata.utc_time or nil
end



function modmm.selectInstallDir(fullmodid)
	local modsdir = modmm.getDirLocalMods()
	if (modsdir == nil) then
		return false, "Can't find local mod dir"
	end

	local modsgamedir = modmm.getDirGameMods()
	if (modsgamedir == nil) then
		return false, "Can't find game mod dir"
	end

	local destdir = modsdir..'/'..fullmodid
	local destdirgame = modsgamedir..'/'..fullmodid

	local destdir_real = _native.fs.realpath(destdir)
	if (destdir_real ~= nil) then
		_log("Found old mod @ "..destdir_real)
	else
		local destdirgame_real = _native.fs.realpath(destdirgame)
		if (destdirgame_real ~= nil) then
			_log("Found old mod @ "..destdirgame_real)
			destdir = destdirgame
		end
	end
	return true, destdir
end

function modmm.downloadAndInstallMod(repoentry)
	local ok, errormsg, destdir
	local downloaddir = modmm.getDirFor("downloads")
	if (downloaddir == nil) then
		return false, "Can't create downloads"
	end

	local modstoragedir = modmm.getDirFor("modstorage")
	if (modstoragedir == nil) then
		return false, "Can't create modstorage"
	end

	if (repoentry.modid == nil) then
		return false, "Internal error in repository, modid not found"
	end

	ok, destdir = modmm.selectInstallDir(repoentry.modid)
	if (ok == false) then
		return false, destdir
	end

	local downloadurl = repoentry.repodata.download_prefix_url..repoentry.download
	local strversion = 	''..table.concat(repoentry.version or '1', '.')
	local downloadextract =  downloaddir..'/'..repoentry.modid..'__'..strversion
	local downloadfile = downloadextract..'.package'

	ok, errormsg = modmm.download(downloadfile, downloadurl)

	if (ok == false) then
		return false, errormsg
	end

	ok, errormsg = modmm._doInstallMod(repoentry.modid, downloadfile, downloadextract, modstoragedir, destdir)

	if (ok == false) then
		return false, errormsg
	end
end


function modmm._doInstallMod(fullmodid, downloadfile, downloadextract, modstoragedir, destdir)
	local ok, errormsg = modmm.extractFile(downloadfile, fullmodid, downloadextract)
	if (ok == false) then
		return false, errormsg
	end

	local destdir_real = _native.fs.realpath(destdir)
	if (destdir_real ~= nil) then
		local savedfolder = modstoragedir..'/'..fullmodid..'-saved-'..os.time()
		_log("Moving old mod folder to new dest:\n"..savedfolder);
		errormsg = _native.fs.rename(destdir_real, savedfolder)
		if (errormsg ~= nil) then
			return false, "Moving old mod folder failed:\n"..errormsg
		end
	end

	_log("Moving extracted folder to new dest:\n"..destdir);
	errormsg = _native.fs.rename(downloadextract..'/'..fullmodid, destdir)
	if (errormsg ~= nil) then
		return false, "Moving extracted folder to new dest failed:\n"..errormsg
	end
	return true
end


function modmm.installModFromArchiv(filename)
	local ok, errormsg, fullmodid, destdir

	local downloaddir = modmm.getDirFor("downloads")
	if (downloaddir == nil) then
		return false, "Can't create downloads"
	end

	local modstoragedir = modmm.getDirFor("modstorage")
	if (modstoragedir == nil) then
		return false, "Can't create modstorage"
	end

	ok, fullmodid = modmm.getFileInfo(filename)
	if (ok == false) then
		return false, fullmodid
	end

	ok, destdir = modmm.selectInstallDir(fullmodid)
	if (ok == false) then
		return false, destdir
	end

	local downloadextract =  downloaddir..'/local-'..fullmodid..'__'..os.time()

	ok, errormsg = modmm._doInstallMod(fullmodid, filename, downloadextract, modstoragedir, destdir)
	if (ok == false) then
		return false, errormsg
	end
	return true
end


function modmm.extractFile(filename, direntry, extractto)
	local pcmd
	local q_file = _utils.quoteParam(filename)
	local q_direntry = _utils.quoteParam(direntry)
	local q_extractto = _utils.quoteParam(extractto)

	local q_cmd = ' x '..q_file..' -o'..q_extractto..' '..q_direntry..' -r -aoa -bd -bso0 -bsp0 -bse1'
	if (commonapi._info.sys == 1) then
		local q_7zip = _utils.quoteParam("C:\\Program Files\\7-Zip\\7z.exe")
		pcmd = '"'..q_7zip..q_cmd..'"'
	else
		local q_7zip = _utils.quoteParam("/usr/bin/7z")
		pcmd = 'LD_LIBRARY_PATH=; LD_PRELOAD=; '..q_7zip..q_cmd
	end

	_log("Trying to run "..pcmd)

	local p = io.popen(pcmd)
	if (p ~= nil) then
		local msg = p:read("*a")
		print(msg)
		p:close()
		if (msg ~= "") then
			return false, msg
		end
	end

	return true
end


function modmm.getFileInfo(filename)
	local pcmd
	local REGEX_MODFOLDER = "^([a-zA-Z0-9_-]+_([0-9]+))%/mod%.lua$"
	local q_file = _utils.quoteParam(filename)
	local q_direntry = _utils.quoteParam('*/mod.lua')

	local q_cmd = ' l -slt -bd -bso0 -bsp0 -bse1 '..q_file..' '..q_direntry
	if (commonapi._info.sys == 1) then
		local q_7zip = _utils.quoteParam("C:\\Program Files\\7-Zip\\7z.exe")
		pcmd = '"'..q_7zip..q_cmd..'"'
	else
		local q_7zip = _utils.quoteParam("/usr/bin/7z")
		pcmd = 'LD_LIBRARY_PATH=; LD_PRELOAD=; '..q_7zip..q_cmd
	end

	_log("Trying to run "..pcmd)

	local modid
	local lines = {}
	local p = io.popen(pcmd)
	if (p ~= nil) then
		for line in p:lines() do
			lines[#lines+1] = line
		end

		p:close()

		msg = table.concat(lines, "\n")
		--commonapi.dmp(lines)
		--print(msg)
		local files = modmm.parse7zipoutput(lines)

		for i, file in ipairs(files) do
			modid, modmajorversion = file:match(REGEX_MODFOLDER)
			if (modid ~= nil) then
				print("Found modid"..modid)
				break;
			end
		end
		if (modid ~= nil) then
			return true, modid
		end

		if (msg ~= "") then
			return false, msg
		end
	end

	return false, "Can't get modid from "..filename
end


function modmm.parse7zipoutput(lines)
	local files = {}
	local startoflist = '----------'
	local pathline = 'Path = '
	local inheader = true
	for i, line in ipairs(lines) do
		if (inheader) then
			if (line == startoflist) then
				inheader = false
			end
		else
			if (line:sub(1, #pathline) == pathline) then
				files[#files+1] = string.gsub( line:sub(#pathline+1), '\\', '/')
			end
		end
	end
	--commonapi.dmp(files)

	return files
end
return modmm