---- Stationhelper interface
-- @author Oskar Eisemuth
-- @copyright 2019
-- @module commonapi.stationhelper
local stationhelper = {
	
}

local _trainstationutil = require "modules/trainstationutil"
local _utils = commonapi.utils

-- print("commonapi2.stationhelper loaded")

local vec3 = require "vec3"
local transf = require "transf"


function stationhelper._makeTrackInternal(result, transform, tag, slotId, addModuleFn, params, edgeListNum)
	if (commonapi._native ~= nil and commonapi._native.renameLuaState ~= nil) then
		commonapi._native.renameLuaState("Construction Thread")
	end


	local pos = vec3.new(transform[13], transform[14], transform[15])
	local edgeList = result.edgeLists[edgeListNum]
	local edges = edgeList.edges
	local nedges = #(edges)

	result.models[#result.models + 1] = { 
		id = "station/rail/era_a/tn_underground_connection.mdl",
		transf = transf.mul(transform, transf.transl(vec3.new(0, 0, 0))),
		tag = tag
	}

	edges[nedges + 1] = { { pos.x, pos.y - 18, .0 }, { .0, -2.0, .0 } }
	edges[nedges + 2] = { { pos.x, pos.y - 20, .0 }, { .0, -2.0, .0 } }

	edges[nedges + 3] = { { pos.x, pos.y - 2, .0 }, { .0, -16.0, .0 } }
	edges[nedges + 4] = { { pos.x, pos.y - 18,  .0 }, { .0, -16.0, .0 } }

	edges[nedges + 5] = { { pos.x, pos.y + 0,  .0 }, { .0, -2.0, .0 } }
	edges[nedges + 6] = { { pos.x, pos.y - 2,  .0 }, { .0, -2.0, .0 } }

	edges[nedges + 7] = { { pos.x, pos.y + 0,  .0 }, { .0, 2.0, .0 } }
	edges[nedges + 8] = { { pos.x, pos.y + 2,  .0 }, { .0, 2.0, .0 } }

	edges[nedges + 9] = { { pos.x, pos.y + 2,  .0 }, { .0, 16.0, .0 } }
	edges[nedges + 10] = { { pos.x, pos.y + 18, .0 }, { .0, 16.0, .0 } }

	edges[nedges + 11] = { { pos.x, pos.y + 18, .0 }, { .0, 2.0, .0 } }
	edges[nedges + 12] = { { pos.x, pos.y + 20, .0 }, { .0, 2.0, .0 } }

	local coords = result.GetCoord(slotId)
	local i = coords[1]
	local j = coords[2]

	result.trackCoord2models[i][j].nodeCenterRight = {edgeListNum, nedges + 10}

	result.trackCoord2models[i][j].nodeRight = {edgeListNum, nedges + 7}
	result.trackCoord2models[i][j].nodeLeft = {edgeListNum, nedges + 5}

	result.trackCoord2models[i][j].nodeCenterLeft = {edgeListNum, nedges + 0}


	local nextM = result.GetModuleAt(i, j-1)
	if not nextM or not nextM.metadata.track then
		table.insert(edgeList.snapNodes, nedges + 1)
	end
	local prevM = result.GetModuleAt(i, j+1)
	if not prevM or not prevM.metadata.track then
		table.insert(edgeList.snapNodes, nedges + 11)
	end


	print("stationhelper._makeTrackInternal called 2")


	local function AddWaitAreaTerminal(i, j, module, leftOrRight)
		local mm = module.metadata
		local model
		local trans

		if leftOrRight then
			-- true = backward = left
			trans = transf.transl(vec3.new(-3.5, 0, 0))
		else
			-- forward
			trans = transf.rotZTransl(math.rad(180), vec3.new(3.5, 0, 0))
		end

		if (mm.common and mm.common.waitarea) then
			print("stationhelper.AddWaitAreaTerminal called")
			local waitarea = mm.common.waitarea
			commonapi.dmp(waitarea)
			if (leftOrRight) then
				if (waitarea.left) then
					model = waitarea.left.model
					trans = waitarea.left.transf
				end
			else
				if (waitarea.right) then
					model = waitarea.right.model
					trans = waitarea.right.transf
				end
			end
		elseif mm.cargo_platform then
			model = "station/rail/era_a/tn_cargo_wait_area.mdl"
		elseif mm.passenger_platform then
			model = "station/rail/era_a/tn_passenger_wait_area.mdl"
		end

		if model then
			print("Adding model: ", model, trans)
			addModuleFn(model, trans)

			if (leftOrRight) then
				result.trackCoord2models[i][j].left = {{#result.models - 1, 0}}
			else
				result.trackCoord2models[i][j].right = {{#result.models - 1, 0}}
			end
		end
	end

	local forwardM = result.GetModuleAt(i+1, j)
	if forwardM then
		AddWaitAreaTerminal(i, j, forwardM, false)
	end
	local backwardM = result.GetModuleAt(i-1, j)
	if backwardM then
		AddWaitAreaTerminal(i, j, backwardM, true)
	end

	if (edgeList.tag2nodes[tag] == nil) then
		edgeList.tag2nodes[tag] = {}
	end

	for nC = 0, 11 do
		table.insert(edgeList.tag2nodes[tag], nedges + nC)
	end
end

stationhelper.createEdgeListForTrackType = function(type, catenary)
	return { 
		type = "TRACK",
		params = {
			type =  type,
			catenary = catenary,
		},
		edges = { },
		snapNodes = { },
		tag2nodes = { },
	}
end


stationhelper.getSumOfEdgesBeforeEdgeList = function(edgeLists, index)
	local cnt = 0
	if (index > 1) then
		for i = 1, index-1 do
			cnt = cnt + #edgeLists[i].edges
		end
	end
	return cnt
end

stationhelper.createAutoSumOfEdgesList = function(edgeLists, cumsum)
	if (cumsum == nil) then
		cumsum = {}
	end
	setmetatable(cumsum, {
		__index = function(t, key)
			local v = stationhelper.getSumOfEdgesBeforeEdgeList(edgeLists, key);
			rawset(t, key, v)
			return v
		end
	})
	return cumsum
end


stationhelper.getEdgeListPositionByTrackType = function(edgeLists, type, trackType, catenary)
--	print ("getEdgeListPositionByTrackType")
	local edgeListNum = 0
	for i = 1, #edgeLists do
		local entry = edgeLists[i]
		if ( (entry.type == type) and (entry.params ~= nil)) then
			if ( (entry.params.type == trackType) and (entry.params.catenary == catenary) ) then
				edgeListNum = i
			end
		end
	end
--	print ("getEdgeListPositionByTrackType edgeListNum:", edgeListNum)
	return edgeListNum
end

stationhelper.getEdgeListPositionByTrackTypeCreate = function(edgeLists, type, trackType, catenary)
	local edgeListNum = stationhelper.getEdgeListPositionByTrackType(edgeLists, type, trackType, catenary)
	if (edgeListNum == 0) then
		edgeListNum = #edgeLists+1
		edgeLists[edgeListNum] = stationhelper.createEdgeListForTrackType(trackType, catenary)
	end
	return edgeListNum
end


stationhelper.makeTrackByTrackType = function(result, transform, tag, slotId, addModuleFn, params, trackType, catenary)
	local edgeListNum = stationhelper.getEdgeListPositionByTrackTypeCreate(result.edgeLists, "TRACK", trackType, catenary)
	stationhelper._makeTrackInternal(result, transform, tag, slotId, addModuleFn, params, edgeListNum)
end

stationhelper.getEdgeListPositionByStreetypeCreate = function(edgeLists, type, trackType, catenary)
	local edgeListNum = stationhelper.getEdgeListPositionByTrackType(edgeLists, type, trackType, catenary)
	if (edgeListNum == 0) then
		edgeListNum = #edgeLists+1
		edgeLists[edgeListNum] = stationhelper.createEdgeListForTrackType(trackType, catenary)
	end
	return edgeListNum
end

stationhelper._defaultModuleData = {
	["station/rail/modular_station/platform_track.module"] = {
		hasCatenary = false,
		isHighSpeed = false,
		toggleCatenary = "station/rail/modular_station/platform_track_catenary.module",
		toggleTrack = "station/rail/modular_station/platform_high_speed_track.module"
	},
	["station/rail/modular_station/platform_track_catenary.module"] = {
		hasCatenary = true,
		isHighSpeed = false,
		toggleCatenary = "station/rail/modular_station/platform_track.module",
		toggleTrack = "station/rail/modular_station/platform_high_speed_track_catenary.module"
	},

	["station/rail/modular_station/platform_high_speed_track.module"] = {
		hasCatenary = false,
		isHighSpeed = true,
		toggleCatenary = "station/rail/modular_station/platform_high_speed_track_catenary.module",
		toggleTrack = "station/rail/modular_station/platform_track.module"
	},
	["station/rail/modular_station/platform_high_speed_track_catenary.module"] = {
		hasCatenary = true,
		isHighSpeed = true,
		toggleCatenary = "station/rail/modular_station/platform_high_speed_track.module",
		toggleTrack = "station/rail/modular_station/platform_track_catenary.module"
	}
}

stationhelper.getModuleCommonData = function(data)
	local datacommon = nil
	if (data.metadata and data.metadata.common) then
		datacommon = data.metadata.common
	elseif (data.name ~= nil) then
		datacommon = stationhelper._defaultModuleData[data.name]
	end
	return datacommon
end

stationhelper.toggleCatenaryRow = function(modules, params, slotId, slotForwardFn, slotBackwardFn)
--	print("stationhelper.toggleCatenaryRow")
	local slotId2 = slotId
	local hasCatenary = false
	local datacommon

	if (params.modules[slotId] == nil) then
		print("stationhelper.toggleCatenaryRow params.modules is nil!, slotId = ", slotId)
		return modules
	end

	datacommon = stationhelper.getModuleCommonData( params.modules[slotId] )


	if (datacommon and datacommon.hasCatenary ~= nil) then
		hasCatenary = (datacommon.hasCatenary == true)
	else
		return modules
	end

--	print("stationhelper.toggleCatenaryRow hasCatenary = ", hasCatenary)
	local SwapModuleCatenary = function(slotId, catenary)
		if (params.modules[slotId]) then
			datacommon = stationhelper.getModuleCommonData( params.modules[slotId] )
			--commonapi.dmp( datacommon )
			if (datacommon ~= nil and datacommon.toggleCatenary and datacommon.hasCatenary == catenary) then
--				print("Modifing "..slotId.."  "..datacommon.toggleCatenary)
				table.insert(modules, {slotId, datacommon.toggleCatenary })
				return true
			end
		end
		return false
	end

	while SwapModuleCatenary(slotId2, hasCatenary) do
		--slotId2 = slotId2 + 10
		slotId2 = slotForwardFn(params, slotId2)
	end
	slotId2 = slotId

	while SwapModuleCatenary(slotId2, hasCatenary) do
		-- slotId2 = slotId2 - 10
		slotId2 = slotBackwardFn(params, slotId2)
	end

	return modules
end



stationhelper.toggleTrackRow = function(modules, params, slotId, slotForwardFn, slotBackwardFn)
--	print("stationhelper.toggleTrackRow")
	local slotId2 = slotId
	local isHighSpeed
	local datacommon

	if (params.modules[slotId] == nil) then
		print("stationhelper.toggleTrackRow params.modules is nil!, slotId = ", slotId)
		return modules
	end

	datacommon = stationhelper.getModuleCommonData( params.modules[slotId] )
	if (datacommon and datacommon.isHighSpeed ~= nil) then
		isHighSpeed = (datacommon.isHighSpeed == true)
	else
		return modules
	end

--	print("stationhelper.toggleTrackRow isHighSpeed = ", isHighSpeed)
	local SwapModuleTrack = function(slotId, isHighSpeed)
		if (params.modules[slotId]) then
			datacommon = stationhelper.getModuleCommonData( params.modules[slotId] )
			--commonapi.dmp( datacommon )
			if (datacommon ~= nil and datacommon.toggleTrack and datacommon.isHighSpeed == isHighSpeed) then
--				print("Modifing "..slotId.."  "..datacommon.toggleTrack)
				table.insert(modules, {slotId, datacommon.toggleTrack })
				return true
			end
		end
		return false
	end

	while SwapModuleTrack(slotId2, isHighSpeed) do
		--slotId2 = slotId2 + 10
		slotId2 = slotForwardFn(params, slotId2)
	end
	slotId2 = slotId

	while SwapModuleTrack(slotId2, isHighSpeed) do
		-- slotId2 = slotId2 - 10
		slotId2 = slotBackwardFn(params, slotId2)
	end

	return modules
end


stationhelper.DefaultTrackInterface = {
	makeTrackByTrackType = stationhelper.makeTrackByTrackType
}

stationhelper.createTrackResultInterface = function(result)
	result.common_trackinterface = _utils.deepcopy(stationhelper.DefaultTrackInterface)
end

return stationhelper